@@ -22,22 +22,25 @@
 
 #include "tilesetdock.h"
 
+#include "actionmanager.h"
 #include "addremovemapobject.h"
-#include "addremovetiles.h"
 #include "addremovetileset.h"
 #include "containerhelpers.h"
 #include "documentmanager.h"
-#include "editterraindialog.h"
 #include "erasetiles.h"
 #include "map.h"
 #include "mapdocument.h"
 #include "mapobject.h"
-#include "movetileset.h"
 #include "objectgroup.h"
 #include "preferences.h"
+#include "replacetileset.h"
+#include "swaptiles.h"
+#include "changemapobject.h"
 #include "terrain.h"
 #include "tile.h"
 #include "tilelayer.h"
+#include "tilesetdocument.h"
+#include "tilesetdocumentsmodel.h"
 #include "tilesetformat.h"
 #include "tilesetmodel.h"
 #include "tilesetview.h"
@@ -47,14 +50,15 @@
 #include "utils.h"
 #include "zoomable.h"
 
-#include <QMimeData>
 #include <QAction>
-#include <QDropEvent>
 #include <QComboBox>
+#include <QDropEvent>
 #include <QFileDialog>
 #include <QHBoxLayout>
 #include <QMenu>
 #include <QMessageBox>
+#include <QMimeData>
+#include <QSettings>
 #include <QSignalMapper>
 #include <QStackedWidget>
 #include <QStylePainter>
@@ -70,46 +74,6 @@ using namespace Tiled::Internal;
 
 namespace {
 
-/**
- * Used for exporting/importing tilesets.
- *
- * @warning Does not work for tilesets that are shared by multiple maps!
- */
-class SetTilesetFileName : public QUndoCommand
-{
-public:
-    SetTilesetFileName(MapDocument *mapDocument,
-                       Tileset *tileset,
-                       const QString &fileName)
-        : mMapDocument(mapDocument)
-        , mTileset(tileset)
-        , mFileName(fileName)
-    {
-        if (fileName.isEmpty())
-            setText(QCoreApplication::translate("Undo Commands",
-                                                "Import Tileset"));
-        else
-            setText(QCoreApplication::translate("Undo Commands",
-                                                "Export Tileset"));
-    }
-
-    void undo() override { swap(); }
-    void redo() override { swap(); }
-
-private:
-    void swap()
-    {
-        QString previousFileName = mTileset->fileName();
-        mMapDocument->setTilesetFileName(mTileset, mFileName);
-        mFileName = previousFileName;
-    }
-
-    MapDocument *mMapDocument;
-    Tileset *mTileset;
-    QString mFileName;
-};
-
-
 class TilesetMenuButton : public QToolButton
 {
 public:
@@ -117,7 +81,7 @@ public:
         : QToolButton(parent)
     {
         setArrowType(Qt::DownArrow);
-        setIconSize(QSize(16, 16));
+        setIconSize(Utils::smallIconSize());
         setPopupMode(QToolButton::InstantPopup);
         setAutoRaise(true);
 
@@ -164,25 +128,6 @@ public:
 };
 
 
-static bool hasTileReferences(MapDocument *mapDocument,
-                              std::function<bool(const Cell &)> condition)
-{
-    for (Layer *layer : mapDocument->map()->layers()) {
-        if (TileLayer *tileLayer = layer->asTileLayer()) {
-            if (tileLayer->hasCell(condition))
-                return true;
-
-        } else if (ObjectGroup *objectGroup = layer->asObjectGroup()) {
-            for (MapObject *object : *objectGroup) {
-                if (condition(object->cell()))
-                    return true;
-            }
-        }
-    }
-
-    return false;
-}
-
 static void removeTileReferences(MapDocument *mapDocument,
                                  std::function<bool(const Cell &)> condition)
 {
@@ -205,39 +150,34 @@ static void removeTileReferences(MapDocument *mapDocument,
 
 } // anonymous namespace
 
-TilesetDock::TilesetDock(QWidget *parent):
-    QDockWidget(parent),
-    mMapDocument(nullptr),
-    mTabBar(new WheelEnabledTabBar),
-    mViewStack(new QStackedWidget),
-    mToolBar(new QToolBar),
-    mCurrentTile(nullptr),
-    mCurrentTiles(nullptr),
-    mNewTileset(new QAction(this)),
-    mImportTileset(new QAction(this)),
-    mExportTileset(new QAction(this)),
-    mPropertiesTileset(new QAction(this)),
-    mDeleteTileset(new QAction(this)),
-    mEditTerrain(new QAction(this)),
-    mAddTiles(new QAction(this)),
-    mRemoveTiles(new QAction(this)),
-    mTilesetMenuButton(new TilesetMenuButton(this)),
-    mTilesetMenu(new QMenu(this)),
-    mTilesetActionGroup(new QActionGroup(this)),
-    mTilesetMenuMapper(nullptr),
-    mEmittingStampCaptured(false),
-    mSynchronizingSelection(false)
+TilesetDock::TilesetDock(QWidget *parent)
+    : QDockWidget(parent)
+    , mMapDocument(nullptr)
+    , mTilesetDocumentsFilterModel(new TilesetDocumentsFilterModel(this))
+    , mTabBar(new WheelEnabledTabBar)
+    , mViewStack(new QStackedWidget)
+    , mToolBar(new QToolBar)
+    , mCurrentTile(nullptr)
+    , mCurrentTiles(nullptr)
+    , mNewTileset(new QAction(this))
+    , mEmbedTileset(new QAction(this))
+    , mExportTileset(new QAction(this))
+    , mEditTileset(new QAction(this))
+    , mDeleteTileset(new QAction(this))
+    , mTilesetMenuButton(new TilesetMenuButton(this))
+    , mTilesetMenu(new QMenu(this))
+    , mTilesetActionGroup(new QActionGroup(this))
+    , mTilesetMenuMapper(nullptr)
+    , mEmittingStampCaptured(false)
+    , mSynchronizingSelection(false)
 {
     setObjectName(QLatin1String("TilesetDock"));
 
-    mTabBar->setMovable(true);
     mTabBar->setUsesScrollButtons(true);
     mTabBar->setExpanding(false);
 
-    connect(mTabBar, SIGNAL(currentChanged(int)),
-            SLOT(updateActions()));
-    connect(mTabBar, SIGNAL(tabMoved(int,int)),
-            this, SLOT(moveTileset(int,int)));
+    connect(mTabBar, &QTabBar::currentChanged, this, &TilesetDock::updateActions);
+    connect(mTabBar, &QTabBar::tabMoved, this, &TilesetDock::onTabMoved);
 
     QWidget *w = new QWidget(this);
 
@@ -248,7 +188,7 @@ TilesetDock::TilesetDock(QWidget *parent):
 
     QVBoxLayout *vertical = new QVBoxLayout(w);
     vertical->setSpacing(0);
-    vertical->setMargin(5);
+    vertical->setMargin(0);
     vertical->addLayout(horizontal);
     vertical->addWidget(mViewStack);
 
@@ -258,52 +198,31 @@ TilesetDock::TilesetDock(QWidget *parent):
     vertical->addLayout(horizontal);
 
     mNewTileset->setIcon(QIcon(QLatin1String(":images/16x16/document-new.png")));
-    mImportTileset->setIcon(QIcon(QLatin1String(":images/16x16/document-import.png")));
+    mEmbedTileset->setIcon(QIcon(QLatin1String(":images/16x16/document-import.png")));
     mExportTileset->setIcon(QIcon(QLatin1String(":images/16x16/document-export.png")));
-    mPropertiesTileset->setIcon(QIcon(QLatin1String(":images/16x16/document-properties.png")));
+    mEditTileset->setIcon(QIcon(QLatin1String(":images/16x16/document-properties.png")));
     mDeleteTileset->setIcon(QIcon(QLatin1String(":images/16x16/edit-delete.png")));
-    mEditTerrain->setIcon(QIcon(QLatin1String(":images/16x16/terrain.png")));
-    mAddTiles->setIcon(QIcon(QLatin1String(":images/16x16/add.png")));
-    mRemoveTiles->setIcon(QIcon(QLatin1String(":images/16x16/remove.png")));
 
     Utils::setThemeIcon(mNewTileset, "document-new");
-    Utils::setThemeIcon(mImportTileset, "document-import");
+    Utils::setThemeIcon(mEmbedTileset, "document-import");
     Utils::setThemeIcon(mExportTileset, "document-export");
-    Utils::setThemeIcon(mPropertiesTileset, "document-properties");
+    Utils::setThemeIcon(mEditTileset, "document-properties");
     Utils::setThemeIcon(mDeleteTileset, "edit-delete");
-    Utils::setThemeIcon(mAddTiles, "add");
-    Utils::setThemeIcon(mRemoveTiles, "remove");
 
-    connect(mNewTileset, SIGNAL(triggered()),
-            SIGNAL(newTileset()));
-    connect(mImportTileset, SIGNAL(triggered()),
-            SLOT(importTileset()));
-    connect(mExportTileset, SIGNAL(triggered()),
-            SLOT(exportTileset()));
-    connect(mPropertiesTileset, SIGNAL(triggered()),
-            SLOT(editTilesetProperties()));
-    connect(mDeleteTileset, SIGNAL(triggered()),
-            SLOT(removeTileset()));
-    connect(mEditTerrain, SIGNAL(triggered()),
-            SLOT(editTerrain()));
-    connect(mAddTiles, SIGNAL(triggered()),
-            SLOT(addTiles()));
-    connect(mRemoveTiles, SIGNAL(triggered()),
-            SLOT(removeTiles()));
+    connect(mNewTileset, SIGNAL(triggered()), SLOT(newTileset()));
+    connect(mEmbedTileset, SIGNAL(triggered()), SLOT(embedTileset()));
+    connect(mExportTileset, SIGNAL(triggered()), SLOT(exportTileset()));
+    connect(mEditTileset, SIGNAL(triggered()), SLOT(editTileset()));
+    connect(mDeleteTileset, SIGNAL(triggered()), SLOT(removeTileset()));
 
     mToolBar->addAction(mNewTileset);
-    mToolBar->setIconSize(QSize(16, 16));
-    mToolBar->addAction(mImportTileset);
+    mToolBar->setIconSize(Utils::smallIconSize());
+    mToolBar->addAction(mEmbedTileset);
     mToolBar->addAction(mExportTileset);
-    mToolBar->addAction(mPropertiesTileset);
+    mToolBar->addAction(mEditTileset);
     mToolBar->addAction(mDeleteTileset);
-    mToolBar->addAction(mEditTerrain);
-    mToolBar->addAction(mAddTiles);
-    mToolBar->addAction(mRemoveTiles);
 
-    mZoomable = new Zoomable(this);
     mZoomComboBox = new QComboBox;
-    mZoomable->connectToComboBox(mZoomComboBox);
     horizontal->addWidget(mZoomComboBox);
 
     connect(mViewStack, &QStackedWidget::currentChanged,
@@ -311,11 +230,19 @@ TilesetDock::TilesetDock(QWidget *parent):
     connect(mViewStack, &QStackedWidget::currentChanged,
             this, &TilesetDock::currentTilesetChanged);
 
-    connect(TilesetManager::instance(), SIGNAL(tilesetChanged(Tileset*)),
-            this, SLOT(tilesetChanged(Tileset*)));
+    connect(TilesetManager::instance(), &TilesetManager::tilesetImagesChanged,
+            this, &TilesetDock::tilesetChanged);
 
-    connect(DocumentManager::instance(), SIGNAL(documentAboutToClose(MapDocument*)),
-            SLOT(documentAboutToClose(MapDocument*)));
+    connect(mTilesetDocumentsFilterModel, &TilesetDocumentsModel::rowsInserted,
+            this, &TilesetDock::onTilesetRowsInserted);
+    connect(mTilesetDocumentsFilterModel, &TilesetDocumentsModel::rowsAboutToBeRemoved,
+            this, &TilesetDock::onTilesetRowsAboutToBeRemoved);
+    connect(mTilesetDocumentsFilterModel, &TilesetDocumentsModel::rowsMoved,
+            this, &TilesetDock::onTilesetRowsMoved);
+    connect(mTilesetDocumentsFilterModel, &TilesetDocumentsModel::layoutChanged,
+            this, &TilesetDock::onTilesetLayoutChanged);
+    connect(mTilesetDocumentsFilterModel, &TilesetDocumentsModel::dataChanged,
+            this, &TilesetDock::onTilesetDataChanged);
 
     mTilesetMenuButton->setMenu(mTilesetMenu);
     connect(mTilesetMenu, SIGNAL(aboutToShow()), SLOT(refreshTilesetMenu()));
@@ -323,6 +250,7 @@ TilesetDock::TilesetDock(QWidget *parent):
     setWidget(w);
     retranslateUi();
     setAcceptDrops(true);
+
     updateActions();
 }
 
@@ -345,68 +273,25 @@ void TilesetDock::setMapDocument(MapDocument *mapDocument)
     setCurrentTiles(nullptr);
     setCurrentTile(nullptr);
 
-    if (mMapDocument) {
-        // Remember the last visible tileset for this map
-        const QString tilesetName = mTabBar->tabText(mTabBar->currentIndex());
-        mCurrentTilesets.insert(mMapDocument, tilesetName);
-    }
-
-    // Clear previous content
-    while (mTabBar->count())
-        mTabBar->removeTab(0);
-    while (mViewStack->count())
-        delete mViewStack->widget(0);
-
-    mTilesets.clear();
-
     // Clear all connections to the previous document
     if (mMapDocument)
         mMapDocument->disconnect(this);
 
     mMapDocument = mapDocument;
 
+    mTilesetDocumentsFilterModel->setMapDocument(mapDocument);
+
     if (mMapDocument) {
-        mTilesets = mMapDocument->map()->tilesets();
-
-        for (const SharedTileset &tileset : mTilesets) {
-            TilesetView *view = new TilesetView;
-            view->setMapDocument(mMapDocument);
-            view->setZoomable(mZoomable);
-
-            mTabBar->addTab(tileset->name());
-            mViewStack->addWidget(view);
-        }
-
-        connect(mMapDocument, &MapDocument::tilesetAdded,
-                this, &TilesetDock::tilesetAdded);
-        connect(mMapDocument, &MapDocument::tilesetRemoved,
-                this, &TilesetDock::tilesetRemoved);
-        connect(mMapDocument, &MapDocument::tilesetMoved,
-                this, &TilesetDock::tilesetMoved);
-        connect(mMapDocument, &MapDocument::tilesetReplaced,
-                this, &TilesetDock::tilesetReplaced);
-        connect(mMapDocument, &MapDocument::tilesetNameChanged,
-                this, &TilesetDock::tilesetNameChanged);
-        connect(mMapDocument, &MapDocument::tilesetFileNameChanged,
-                this, &TilesetDock::updateActions);
-        connect(mMapDocument, &MapDocument::tilesetChanged,
-                this, &TilesetDock::tilesetChanged);
-        connect(mMapDocument, &MapDocument::tileImageSourceChanged,
-                this, &TilesetDock::tileImageSourceChanged);
-        connect(mMapDocument, &MapDocument::tileAnimationChanged,
-                this, &TilesetDock::tileAnimationChanged);
-
-        QString cacheName = mCurrentTilesets.take(mMapDocument);
-        for (int i = 0; i < mTabBar->count(); ++i) {
-            if (mTabBar->tabText(i) == cacheName) {
-                mTabBar->setCurrentIndex(i);
-                break;
-            }
-        }
-
         if (Object *object = mMapDocument->currentObject())
             if (object->typeId() == Object::TileType)
                 setCurrentTile(static_cast<Tile*>(object));
+
+        connect(mMapDocument, &MapDocument::tilesetAdded,
+                this, &TilesetDock::updateActions);
+        connect(mMapDocument, &MapDocument::tilesetRemoved,
+                this, &TilesetDock::updateActions);
+        connect(mMapDocument, &MapDocument::tilesetReplaced,
+                this, &TilesetDock::updateActions);
     }
 
     updateActions();
@@ -431,7 +316,7 @@ void TilesetDock::selectTilesInStamp(const TileStamp &stamp)
     for (const TileStampVariation &variation : stamp.variations()) {
         const TileLayer &tileLayer = *variation.tileLayer();
         for (const Cell &cell : tileLayer) {
-            if (Tile *tile = cell.tile) {
+            if (Tile *tile = cell.tile()) {
                 if (processed.contains(tile))
                     continue;
 
@@ -463,9 +348,6 @@ void TilesetDock::selectTilesInStamp(const TileStamp &stamp)
             selectionModel->select(selection, QItemSelectionModel::SelectCurrent);
         }
 
-        // Show/edit properties of all captured tiles
-        mMapDocument->setSelectedTiles(processed.toList());
-
         // Update the current tile (useful for animation and collision editors)
         auto first = selections.begin();
         QItemSelectionModel *selectionModel = first.key();
@@ -496,7 +378,7 @@ void TilesetDock::dragEnterEvent(QDragEnterEvent *e)
 {
     const QList<QUrl> urls = e->mimeData()->urls();
     if (!urls.isEmpty() && !urls.at(0).toLocalFile().isEmpty())
-        e->accept();
+        e->acceptProposedAction();
 }
 
 void TilesetDock::dropEvent(QDropEvent *e)
@@ -508,16 +390,19 @@ void TilesetDock::dropEvent(QDropEvent *e)
             paths.append(localFile);
     }
     if (!paths.isEmpty()) {
-        emit tilesetsDropped(paths);
-        e->accept();
+        emit localFilesDropped(paths);
+        e->acceptProposedAction();
     }
 }
 
 void TilesetDock::currentTilesetChanged()
 {
-    if (const TilesetView *view = currentTilesetView())
+    if (const TilesetView *view = currentTilesetView()) {
+        view->zoomable()->setComboBox(mZoomComboBox);
+
         if (const QItemSelectionModel *s = view->selectionModel())
             setCurrentTile(view->tilesetModel()->tileAt(s->currentIndex()));
+    }
 }
 
 void TilesetDock::selectionChanged()
@@ -540,38 +425,30 @@ void TilesetDock::currentChanged(const QModelIndex &index)
 void TilesetDock::updateActions()
 {
     bool external = false;
-    bool isCollection = false;
-    bool hasSelection = false;
     TilesetView *view = nullptr;
+    Tileset *tileset = nullptr;
     const int index = mTabBar->currentIndex();
 
     if (index > -1) {
         view = tilesetViewAt(index);
-        if (view) {
-            Tileset *tileset = mTilesets.at(index).data();
+        tileset = mTilesets.at(index).data();
 
+        if (view) {
             if (!view->model()) // Lazily set up the model
                 setupTilesetModel(view, tileset);
 
             mViewStack->setCurrentIndex(index);
             external = tileset->isExternal();
-            isCollection = tileset->isCollection();
-            hasSelection = view->selectionModel()->hasSelection();
         }
     }
 
     const bool tilesetIsDisplayed = view != nullptr;
-    const bool mapIsDisplayed = mMapDocument != nullptr;
+    const auto map = mMapDocument ? mMapDocument->map() : nullptr;
 
-    mNewTileset->setEnabled(mapIsDisplayed);
-    mImportTileset->setEnabled(tilesetIsDisplayed && external);
+    mEmbedTileset->setEnabled(tilesetIsDisplayed && external);
     mExportTileset->setEnabled(tilesetIsDisplayed && !external);
-    mPropertiesTileset->setEnabled(tilesetIsDisplayed);
-    mDeleteTileset->setEnabled(tilesetIsDisplayed);
-    mEditTerrain->setEnabled(tilesetIsDisplayed && !external);
-    mAddTiles->setEnabled(tilesetIsDisplayed && isCollection && !external);
-    mRemoveTiles->setEnabled(tilesetIsDisplayed && isCollection
-                             && hasSelection && !external);
+    mEditTileset->setEnabled(tilesetIsDisplayed);
+    mDeleteTileset->setEnabled(tilesetIsDisplayed && map && contains(map->tilesets(), tileset));
 }
 
 void TilesetDock::updateCurrentTiles()
@@ -623,17 +500,75 @@ void TilesetDock::indexPressed(const QModelIndex &index)
         mMapDocument->setCurrentObject(tile);
 }
 
-void TilesetDock::tilesetAdded(int index, Tileset *tileset)
+void TilesetDock::createTilesetView(int index, TilesetDocument *tilesetDocument)
 {
+    auto tileset = tilesetDocument->tileset();
+
+    mTilesets.insert(index, tileset);
+    mTilesetDocuments.insert(index, tilesetDocument);
+
     TilesetView *view = new TilesetView;
-    view->setMapDocument(mMapDocument);
-    view->setZoomable(mZoomable);
 
-    mTilesets.insert(index, tileset->sharedPointer());
-    mTabBar->insertTab(index, tileset->name());
+    // Insert view before the tab to make sure it is there when the tab index
+    // changes (happens when first tab is inserted).
     mViewStack->insertWidget(index, view);
+    mTabBar->insertTab(index, tileset->name());
+    mTabBar->setTabToolTip(index, tileset->fileName());
 
-    updateActions();
+    QString path = QLatin1String("TilesetDock/TilesetScale/") + tileset->name();
+    qreal scale = Preferences::instance()->settings()->value(path, 1).toReal();
+    view->zoomable()->setScale(scale);
+
+    connect(tilesetDocument, &TilesetDocument::fileNameChanged,
+            this, &TilesetDock::tilesetFileNameChanged);
+    connect(tilesetDocument, &TilesetDocument::tilesetChanged,
+            this, &TilesetDock::tilesetChanged);
+    connect(tilesetDocument, &TilesetDocument::tileImageSourceChanged,
+            this, &TilesetDock::tileImageSourceChanged);
+    connect(tilesetDocument, &TilesetDocument::tileAnimationChanged,
+            this, &TilesetDock::tileAnimationChanged);
+
+    connect(view, &TilesetView::clicked,
+            this, &TilesetDock::updateCurrentTiles);
+    connect(view, &TilesetView::swapTilesRequested,
+            this, &TilesetDock::swapTiles);
+    connect(view, &TilesetView::changeSelectedMapObjectsTileRequested,
+            this, &TilesetDock::changeSelectedMapObjectsTile);
+}
+
+void TilesetDock::deleteTilesetView(int index)
+{
+    TilesetDocument *tilesetDocument = mTilesetDocuments.at(index);
+    tilesetDocument->disconnect(this);
+
+    Tileset *tileset = tilesetDocument->tileset().data();
+    TilesetView *view = tilesetViewAt(index);
+
+    QString path = QLatin1String("TilesetDock/TilesetScale/") + tileset->name();
+    QSettings *settings = Preferences::instance()->settings();
+    if (view->scale() != 1.0)
+        settings->setValue(path, view->scale());
+    else
+        settings->remove(path);
+
+    mTilesets.remove(index);
+    mTilesetDocuments.removeAt(index);
+    delete view;                    // view needs to go before the tab
+    mTabBar->removeTab(index);
+
+    // Make sure we don't reference this tileset anymore
+    if (mCurrentTiles && mCurrentTiles->referencesTileset(tileset)) {
+        TileLayer *cleaned = mCurrentTiles->clone();
+        cleaned->removeReferencesToTileset(tileset);
+        setCurrentTiles(cleaned);
+    }
+    if (mCurrentTile && mCurrentTile->tileset() == tileset)
+        setCurrentTile(nullptr);
+}
+
+void TilesetDock::moveTilesetView(int from, int to)
+{
+    mTabBar->moveTab(from, to);
 }
 
 void TilesetDock::tilesetChanged(Tileset *tileset)
@@ -643,67 +578,12 @@ void TilesetDock::tilesetChanged(Tileset *tileset)
     if (index < 0)
         return;
 
-    if (TilesetModel *model = tilesetViewAt(index)->tilesetModel())
+    TilesetView *view = tilesetViewAt(index);
+
+    if (TilesetModel *model = view->tilesetModel()) {
+        view->updateBackgroundColor();
         model->tilesetChanged();
-}
-
-void TilesetDock::tilesetRemoved(Tileset *tileset)
-{
-    // Delete the related tileset view
-    const int index = indexOf(mTilesets, tileset);
-    Q_ASSERT(index != -1);
-
-    mTilesets.remove(index);
-    mTabBar->removeTab(index);
-    delete tilesetViewAt(index);
-
-    // Make sure we don't reference this tileset anymore
-    if (mCurrentTiles) {
-        // TODO: Don't clean unnecessarily (but first the concept of
-        //       "current brush" would need to be introduced)
-        TileLayer *cleaned = static_cast<TileLayer *>(mCurrentTiles->clone());
-        cleaned->removeReferencesToTileset(tileset);
-        setCurrentTiles(cleaned);
     }
-    if (mCurrentTile && mCurrentTile->tileset() == tileset)
-        setCurrentTile(nullptr);
-
-    updateActions();
-}
-
-void TilesetDock::tilesetMoved(int from, int to)
-{
-#if QT_VERSION >= 0x050600
-    mTilesets.move(from, to);
-#else
-    mTilesets.insert(to, mTilesets.takeAt(from));
-#endif
-
-    // Move the related tileset views
-    QWidget *widget = mViewStack->widget(from);
-    mViewStack->removeWidget(widget);
-    mViewStack->insertWidget(to, widget);
-    mViewStack->setCurrentIndex(mTabBar->currentIndex());
-
-    // Update the titles of the affected tabs
-    const int start = qMin(from, to);
-    const int end = qMax(from, to);
-    for (int i = start; i <= end; ++i) {
-        const SharedTileset &tileset = mTilesets.at(i);
-        if (mTabBar->tabText(i) != tileset->name())
-            mTabBar->setTabText(i, tileset->name());
-    }
-}
-
-void TilesetDock::tilesetReplaced(int index, Tileset *tileset)
-{
-    mTilesets.replace(index, tileset->sharedPointer());
-
-    if (TilesetModel *model = tilesetViewAt(index)->tilesetModel())
-        model->setTileset(tileset);
-
-    if (mTabBar->tabText(index) != tileset->name())
-        mTabBar->setTabText(index, tileset->name());
 }
 
 /**
@@ -722,7 +602,13 @@ void TilesetDock::removeTileset()
  */
 void TilesetDock::removeTileset(int index)
 {
-    Tileset *tileset = mTilesets.at(index).data();
+    auto &sharedTileset = mTilesets.at(index);
+
+    int mapTilesetIndex = mMapDocument->map()->tilesets().indexOf(sharedTileset);
+    if (mapTilesetIndex == -1)
+        return;
+
+    Tileset *tileset = sharedTileset.data();
     const bool inUse = mMapDocument->map()->isTilesetUsed(tileset);
 
     // If the tileset is in use, warn the user and confirm removal
@@ -741,15 +627,13 @@ void TilesetDock::removeTileset(int index)
             return;
     }
 
-    QUndoCommand *remove = new RemoveTileset(mMapDocument, index);
+    QUndoCommand *remove = new RemoveTileset(mMapDocument, mapTilesetIndex);
     QUndoStack *undoStack = mMapDocument->undoStack();
 
     if (inUse) {
         // Remove references to tiles in this tileset from the current map
         auto referencesTileset = [tileset] (const Cell &cell) {
-            if (const Tile *tile = cell.tile)
-                return tile->tileset() == tileset;
-            return false;
+            return cell.tileset() == tileset;
         };
         undoStack->beginMacro(remove->text());
         removeTileReferences(mMapDocument, referencesTileset);
@@ -759,10 +643,9 @@ void TilesetDock::removeTileset(int index)
         undoStack->endMacro();
 }
 
-void TilesetDock::moveTileset(int from, int to)
+void TilesetDock::newTileset()
 {
-    QUndoCommand *command = new MoveTileset(mMapDocument, from, to);
-    mMapDocument->undoStack()->push(command);
+    ActionManager::action("file.new_tileset")->trigger();
 }
 
 void TilesetDock::setCurrentTiles(TileLayer *tiles)
@@ -773,18 +656,7 @@ void TilesetDock::setCurrentTiles(TileLayer *tiles)
     delete mCurrentTiles;
     mCurrentTiles = tiles;
 
-    // Set the selected tiles on the map document
-    if (tiles) {
-        QList<Tile*> selectedTiles;
-        for (int y = 0; y < tiles->height(); ++y) {
-            for (int x = 0; x < tiles->width(); ++x) {
-                const Cell &cell = tiles->cellAt(x, y);
-                if (!cell.isEmpty())
-                    selectedTiles.append(cell.tile);
-            }
-        }
-        mMapDocument->setSelectedTiles(selectedTiles);
-
+    if (tiles && mMapDocument) {
         // Create a tile stamp with these tiles
         Map *map = mMapDocument->map();
         Map *stamp = new Map(map->orientation(),
@@ -809,7 +681,7 @@ void TilesetDock::setCurrentTile(Tile *tile)
     mCurrentTile = tile;
     emit currentTileChanged(tile);
 
-    if (tile)
+    if (mMapDocument && tile)
         mMapDocument->setCurrentObject(tile);
 }
 
@@ -817,13 +689,92 @@ void TilesetDock::retranslateUi()
 {
     setWindowTitle(tr("Tilesets"));
     mNewTileset->setText(tr("New Tileset"));
-    mImportTileset->setText(tr("&Import Tileset"));
+    mEmbedTileset->setText(tr("&Embed Tileset"));
     mExportTileset->setText(tr("&Export Tileset As..."));
-    mPropertiesTileset->setText(tr("Tile&set Properties"));
+    mEditTileset->setText(tr("Edit Tile&set"));
     mDeleteTileset->setText(tr("&Remove Tileset"));
-    mEditTerrain->setText(tr("Edit &Terrain Information"));
-    mAddTiles->setText(tr("Add Tiles"));
-    mRemoveTiles->setText(tr("Remove Tiles"));
+}
+
+void TilesetDock::onTilesetRowsInserted(const QModelIndex &parent, int first, int last)
+{
+    for (int row = first; row <= last; ++row) {
+        const QModelIndex index = mTilesetDocumentsFilterModel->index(row, 0, parent);
+        const QVariant var = mTilesetDocumentsFilterModel->data(index, TilesetDocumentsModel::TilesetDocumentRole);
+        createTilesetView(row, var.value<TilesetDocument*>());
+    }
+}
+
+void TilesetDock::onTilesetRowsAboutToBeRemoved(const QModelIndex &parent, int first, int last)
+{
+    Q_UNUSED(parent)
+
+    for (int index = last; index >= first; --index)
+        deleteTilesetView(index);
+}
+
+void TilesetDock::onTilesetRowsMoved(const QModelIndex &parent, int start, int end, const QModelIndex &destination, int row)
+{
+    Q_UNUSED(parent)
+    Q_UNUSED(destination)
+
+    if (start == row)
+        return;
+
+    while (start <= end) {
+        moveTilesetView(start, row);
+
+        if (row < start) {
+            ++start;
+            ++row;
+        } else {
+            --end;
+        }
+    }
+}
+
+void TilesetDock::onTilesetLayoutChanged(const QList<QPersistentModelIndex> &parents, QAbstractItemModel::LayoutChangeHint hint)
+{
+    Q_UNUSED(parents)
+    Q_UNUSED(hint)
+
+    // Make sure the tileset tabs and views are still in the right order
+    for (int i = 0, rows = mTilesetDocumentsFilterModel->rowCount(); i < rows; ++i) {
+        const QModelIndex index = mTilesetDocumentsFilterModel->index(i, 0);
+        const QVariant var = mTilesetDocumentsFilterModel->data(index, TilesetDocumentsModel::TilesetDocumentRole);
+        TilesetDocument *tilesetDocument = var.value<TilesetDocument*>();
+        int currentIndex = mTilesetDocuments.indexOf(tilesetDocument);
+        if (currentIndex != i) {
+            Q_ASSERT(currentIndex > i);
+            moveTilesetView(currentIndex, i);
+        }
+    }
+}
+
+void TilesetDock::onTilesetDataChanged(const QModelIndex &topLeft, const QModelIndex &bottomRight)
+{
+    // Update the titles of the affected tabs
+    for (int i = topLeft.row(); i <= bottomRight.row(); ++i) {
+        const SharedTileset &tileset = mTilesets.at(i);
+        if (mTabBar->tabText(i) != tileset->name())
+            mTabBar->setTabText(i, tileset->name());
+        mTabBar->setTabToolTip(i, tileset->fileName());
+    }
+}
+
+void TilesetDock::onTabMoved(int from, int to)
+{
+#if QT_VERSION >= 0x050600
+    mTilesets.move(from, to);
+#else
+    mTilesets.insert(to, mTilesets.takeAt(from));
+#endif
+    mTilesetDocuments.move(from, to);
+
+    // Move the related tileset view
+    const QSignalBlocker blocker(mViewStack);
+    QWidget *widget = mViewStack->widget(from);
+    mViewStack->removeWidget(widget);
+    mViewStack->insertWidget(to, widget);
 }
 
 Tileset *TilesetDock::currentTileset() const
@@ -858,14 +809,14 @@ void TilesetDock::setupTilesetModel(TilesetView *view, Tileset *tileset)
             this, &TilesetDock::indexPressed);
 }
 
-void TilesetDock::editTilesetProperties()
+void TilesetDock::editTileset()
 {
     Tileset *tileset = currentTileset();
     if (!tileset)
         return;
 
-    mMapDocument->setCurrentObject(tileset);
-    mMapDocument->emitEditCurrentObject();
+    DocumentManager *documentManager = DocumentManager::instance();
+    documentManager->openTileset(tileset->sharedPointer());
 }
 
 void TilesetDock::exportTileset()
@@ -874,21 +825,32 @@ void TilesetDock::exportTileset()
     if (!tileset)
         return;
 
-    const QString tsxFilter = tr("Tiled tileset files (*.tsx)");
+    if (tileset->isExternal())
+        return;
 
-    FormatHelper<TilesetFormat> helper(FileFormat::ReadWrite, tsxFilter);
+    int mapTilesetIndex = mMapDocument->map()->tilesets().indexOf(tileset->sharedPointer());
+    if (mapTilesetIndex == -1)
+        return;
+
+    // To export a tileset we clone it, since the tileset could now be used by
+    // other maps. This ensures undo can take the map back to using an embedded
+    // tileset, without affecting those other maps.
+    SharedTileset externalTileset = tileset->clone();
+
+    FormatHelper<TilesetFormat> helper(FileFormat::ReadWrite);
 
     Preferences *prefs = Preferences::instance();
 
     QString suggestedFileName = prefs->lastPath(Preferences::ExternalTileset);
     suggestedFileName += QLatin1Char('/');
-    suggestedFileName += tileset->name();
+    suggestedFileName += externalTileset->name();
 
     const QLatin1String extension(".tsx");
     if (!suggestedFileName.endsWith(extension))
         suggestedFileName.append(extension);
 
-    QString selectedFilter = tsxFilter;
+    // todo: remember last used filter
+    QString selectedFilter = TsxTilesetFormat().nameFilter();
     const QString fileName =
             QFileDialog::getSaveFileName(this, tr("Export Tileset"),
                                          suggestedFileName,
@@ -900,162 +862,71 @@ void TilesetDock::exportTileset()
     prefs->setLastPath(Preferences::ExternalTileset,
                        QFileInfo(fileName).path());
 
-    TsxTilesetFormat tsxFormat;
     TilesetFormat *format = helper.formatByNameFilter(selectedFilter);
     if (!format)
-        format = &tsxFormat;
+        return;     // can't happen
 
-    if (format->write(*tileset, fileName)) {
-        QUndoCommand *command = new SetTilesetFileName(mMapDocument,
-                                                       tileset,
-                                                       fileName);
-        mMapDocument->undoStack()->push(command);
-    } else {
+    if (!format->write(*externalTileset, fileName)) {
         QString error = format->errorString();
         QMessageBox::critical(window(),
                               tr("Export Tileset"),
                               tr("Error saving tileset: %1").arg(error));
-    }
-}
-
-void TilesetDock::importTileset()
-{
-    Tileset *tileset = currentTileset();
-    if (!tileset)
         return;
+    }
 
-    QUndoCommand *command = new SetTilesetFileName(mMapDocument,
-                                                   tileset,
-                                                   QString());
+    externalTileset->setFileName(fileName);
+    externalTileset->setFormat(format);
+
+    QUndoCommand *command = new ReplaceTileset(mMapDocument,
+                                               mapTilesetIndex,
+                                               externalTileset);
     mMapDocument->undoStack()->push(command);
+
+    // Make sure the external tileset is selected
+    int externalTilesetIndex = mTilesets.indexOf(externalTileset);
+    if (externalTilesetIndex != -1)
+        mTabBar->setCurrentIndex(externalTilesetIndex);
 }
 
-void TilesetDock::editTerrain()
+void TilesetDock::embedTileset()
 {
     Tileset *tileset = currentTileset();
     if (!tileset)
         return;
 
-    EditTerrainDialog editTerrainDialog(mMapDocument, tileset, this);
-    editTerrainDialog.exec();
-}
-
-void TilesetDock::addTiles()
-{
-    Tileset *tileset = currentTileset();
-    if (!tileset)
+    if (!tileset->isExternal())
         return;
 
-    Preferences *prefs = Preferences::instance();
-    const QString startLocation = QFileInfo(prefs->lastPath(Preferences::ImageFile)).absolutePath();
-    const QString filter = Utils::readableImageFormatsFilter();
-    const QStringList files = QFileDialog::getOpenFileNames(window(),
-                                                            tr("Add Tiles"),
-                                                            startLocation,
-                                                            filter);
-    struct LoadedFile {
-        QString imageSource;
-        QPixmap image;
-    };
-    QVector<LoadedFile> loadedFiles;
-
-    for (const QString &file : files) {
-        const QPixmap image(file);
-        if (!image.isNull()) {
-            loadedFiles.append(LoadedFile { file, image });
-        } else {
-            QMessageBox warning(QMessageBox::Warning,
-                                tr("Add Tiles"),
-                                tr("Could not load \"%1\"!").arg(file),
-                                QMessageBox::Ignore | QMessageBox::Cancel,
-                                window());
-            warning.setDefaultButton(QMessageBox::Ignore);
-
-            if (warning.exec() != QMessageBox::Ignore)
-                return;
-        }
-    }
-
-    if (loadedFiles.isEmpty())
-        return;
-
-    prefs->setLastPath(Preferences::ImageFile, files.last());
-
-    QList<Tile*> tiles;
-    tiles.reserve(loadedFiles.size());
-
-    for (LoadedFile &loadedFile : loadedFiles) {
-        Tile *newTile = new Tile(tileset->takeNextTileId(), tileset);
-        newTile->setImage(loadedFile.image);
-        newTile->setImageSource(loadedFile.imageSource);
-        tiles.append(newTile);
-    }
-
-    mMapDocument->undoStack()->push(new AddTiles(mMapDocument,
-                                                 tileset,
-                                                 tiles));
-}
-
-void TilesetDock::removeTiles()
-{
-    TilesetView *view = currentTilesetView();
-    if (!view)
-        return;
-    if (!view->selectionModel()->hasSelection())
-        return;
-
-    const QModelIndexList indexes = view->selectionModel()->selectedIndexes();
-    const TilesetModel *model = view->tilesetModel();
-    QList<Tile*> tiles;
-
-    for (const QModelIndex &index : indexes)
-        if (Tile *tile = model->tileAt(index))
-            tiles.append(tile);
-
-    auto matchesAnyTile = [&tiles] (const Cell &cell) {
-        if (Tile *tile = cell.tile)
-            return tiles.contains(tile);
-        return false;
-    };
-    const bool inUse = hasTileReferences(mMapDocument, matchesAnyTile);
-
-    // If the tileset is in use, warn the user and confirm removal
-    if (inUse) {
-        QMessageBox warning(QMessageBox::Warning,
-                            tr("Remove Tiles"),
-                            tr("One or more of the tiles to be removed are "
-                               "still in use by the map!"),
-                            QMessageBox::Yes | QMessageBox::No,
-                            this);
-        warning.setDefaultButton(QMessageBox::Yes);
-        warning.setInformativeText(tr("Remove all references to these tiles?"));
-
-        if (warning.exec() != QMessageBox::Yes)
-            return;
-    }
+    // To embed a tileset we clone it, since further changes to that tileset
+    // should not affect the exteral tileset.
+    SharedTileset embeddedTileset = tileset->clone();
 
     QUndoStack *undoStack = mMapDocument->undoStack();
-    undoStack->beginMacro(tr("Remove Tiles"));
+    int mapTilesetIndex = mMapDocument->map()->tilesets().indexOf(tileset->sharedPointer());
 
-    if (inUse)
-        removeTileReferences(mMapDocument, matchesAnyTile);
+    // Tileset may not be part of the map yet
+    if (mapTilesetIndex == -1)
+        undoStack->push(new AddTileset(mMapDocument, embeddedTileset));
+    else
+        undoStack->push(new ReplaceTileset(mMapDocument, mapTilesetIndex, embeddedTileset));
 
-    Tileset *tileset = view->tilesetModel()->tileset();
-    undoStack->push(new RemoveTiles(mMapDocument, tileset, tiles));
-
-    undoStack->endMacro();
-
-    // Clear the current tiles, will be referencing the removed tiles
-    setCurrentTiles(nullptr);
-    setCurrentTile(nullptr);
+    // Make sure the embedded tileset is selected
+    int embeddedTilesetIndex = mTilesets.indexOf(embeddedTileset);
+    if (embeddedTilesetIndex != -1)
+        mTabBar->setCurrentIndex(embeddedTilesetIndex);
 }
 
-void TilesetDock::tilesetNameChanged(Tileset *tileset)
+void TilesetDock::tilesetFileNameChanged(const QString &fileName)
 {
+    TilesetDocument *tilesetDocument = static_cast<TilesetDocument*>(sender());
+    Tileset *tileset = tilesetDocument->tileset().data();
+
     const int index = indexOf(mTilesets, tileset);
     Q_ASSERT(index != -1);
 
-    mTabBar->setTabText(index, tileset->name());
+    mTabBar->setTabToolTip(index, fileName);
+
+    updateActions();
 }
 
 void TilesetDock::tileImageSourceChanged(Tile *tile)
@@ -1075,11 +946,6 @@ void TilesetDock::tileAnimationChanged(Tile *tile)
             model->tileChanged(tile);
 }
 
-void TilesetDock::documentAboutToClose(MapDocument *mapDocument)
-{
-    mCurrentTilesets.remove(mapDocument);
-}
-
 void TilesetDock::refreshTilesetMenu()
 {
     mTilesetMenu->clear();
@@ -1108,3 +974,38 @@ void TilesetDock::refreshTilesetMenu()
         mTilesetMenuMapper->setMapping(action, i);
     }
 }
+
+void TilesetDock::swapTiles(Tile *tileA, Tile *tileB)
+{
+    if (!mMapDocument)
+        return;
+
+    QUndoStack *undoStack = mMapDocument->undoStack();
+    undoStack->push(new SwapTiles(mMapDocument, tileA, tileB));
+}
+
+static QList<MapObject*> chooseTileObjects(QList<MapObject*> objects)
+{
+    QList<MapObject*> tileObjects;
+
+    for (auto object : objects)
+        if (!object->cell().isEmpty())
+            tileObjects.append(object);
+
+    return tileObjects;
+}
+
+void TilesetDock::changeSelectedMapObjectsTile(Tile *tile)
+{
+    if (!mMapDocument)
+        return;
+
+    // Only change tiles of tile objects
+    QList<MapObject*> tileObjects = chooseTileObjects(mMapDocument->selectedObjects());
+
+    if (tileObjects.isEmpty())
+        return;
+
+    QUndoStack *undoStack = mMapDocument->undoStack();
+    undoStack->push(new ChangeMapObjectsTile(mMapDocument, tileObjects, tile));
+}
